# -*- coding: utf-8 -*-
"""complex_adam.ipynb

Automatically generated by Colaboratory.

Original file is located at
    https://colab.research.google.com/drive/1M0rrGINzatOuspTwA0TaKfpyrcU2HRMO
"""

import numpy as np
# import jax.numpy as np
import matplotlib.pyplot as plt


def compute_loss(x, y):
  return x * y


def compute_gradient(x, y):
  return np.array([-y, 2 * x])
  # return np.array([2*x, 2*y])


def experiment(step_size, beta_1,
               beta_2=.99 * np.exp(0 * 1j * np.pi), epsilon=1e-8 * np.exp(0),
               do_plot=False):
  num_iterations = 100

  # Initialize some objects
  m = np.array([0 + 0j, 0 + 0j])
  v = np.array([0 + 0j, 0 + 0j])
  x_0, y_0 = 2, 1
  w = np.array([x_0, y_0])
  ws = np.array([w] * num_iterations)
  ws[1:] = ws[1:] * 10e32
  count = 0

  do_synthetic = True  # Set true if we should do extragradient
  for t in range(1, num_iterations):
    # Calculate the extragradient step - denoted synthetic_w
    if do_synthetic:
      synthetic_g = compute_gradient(w[0], w[1])
      synthetic_m = beta_1 * m + synthetic_g  # (1 - beta_1) * g

      synthetic_m_hat = synthetic_m  # / (1 - np.power(beta_1, t))
      synthetic_update = step_size * synthetic_m_hat
      synthetic_update_conj = np.conj(step_size) * np.conj(
        synthetic_m_hat)  # / (np.conj(np.sqrt(v_hat)) + np.conj(epsilon))
      synthetic_combined_update = 0.5 * (synthetic_update + synthetic_update_conj)
      synthetic_w = w - synthetic_combined_update
    else:
      synthetic_w = w

    g = compute_gradient(synthetic_w[0], synthetic_w[1])
    m = beta_1 * m + g  # (1 - beta_1) * g
    m_hat = m  # / (1 - np.power(beta_1, t))
    update = step_size * m_hat
    update_conj = np.conj(step_size) * np.conj(m_hat)  # / (np.conj(np.sqrt(v_hat)) + np.conj(epsilon))
    combined_update = 0.5 * (update + update_conj)
    w = w - combined_update
    ws[t] = np.real(w)  # The imag component is 0, but this gets rid of the imag axis
    if ws[t][0] > 3 or ws[t][0] < -3 or ws[t][1] > 3 or ws[t][1] < -3:
      # Break if we blow up to save compute time in grid search
      break
    # TODO: Break when within a tolerance of the solution
  ws = np.array(ws[:t])

  # This code block calculates ratios for rate of convergence
  ratios = [10e32 for _ in range(1, len(ws))]
  norm = None
  for i in range(0, len(ws)):
    diff = ws[i] - [0, 0]
    old_norm = norm
    norm = diff @ diff.T
    if i > 0:
      if norm < 1e-16:
        ratio = 0
      else:
        if old_norm < 1e-16:
          ratio = 1e16
        else:
          ratio = norm / old_norm
      ratios[i - 1] = ratio

  # This code block does plotting to look at what happens
  if do_plot:
    fig, ax = plt.subplots()

    ax.scatter([x_0], [y_0], label='Init', c='b')
    start_trajectory = 0
    ax.plot(ws[start_trajectory:, 0],
            ws[start_trajectory:, 1],
            label='Trajectory', alpha=.5, c='b')
    ax.scatter([0], [0], label='Solution', c='r')

    circle1 = plt.Circle((0, 0), 1.0, color='g', alpha=.25)
    ax.add_artist(circle1)

    ax.set_xlim([-3, 3])
    ax.set_ylim([-3, 3])
    plt.legend()

    fig, ax = plt.subplots()
    ax.plot(range(1, len(ws)), ratios)
    ax.plot(range(1, len(ws)), [np.mean(ratios[np.maximum(i - 10, 0):i]) for i in range(1, len(ws))])
    ax.set_ylim([-.1, 1.25])

  # Returned a smoothed ratio incase its oscillating
  if len(ratios) == 0:
    return 0, ws
  elif len(ratios) == 1:
    return ratios[0], ws
  return np.mean(ratios[-np.minimum(50, len(ratios)):-1]), ws


# step_size = 0.15789473684210525+0.36842105263157876j
# beta_1 = 0.1578947368421053+0.1578947368421053j
# step_size = .1
# beta_1 = -.17513  # Im(alpha) * Im(beta) = .17513

# step_size = 1+3.551282051282051j
# beta_1 = 0.13793103448275862j
# rate = .54

# EG version
# step_size = 0.8611111111111112+3.3684210526315788j
# beta_1 = 0.07360293770407991+0.10526315789473684j

# step_size = 0.4877551020408163+0.5265306122448979j
# beta_1 = 0.308712060182937+0.21052631578947367j

# NON-EG
# step_size = 0.86111+3.4482j
# beta_1 = 0.07207+0.1207j
# rate = .537

# step_size = .5+.25j  #.5*(1.0 + .5j)
# beta_1 = 0.7666666666666666j
# beta_1 = 0.3230756118960939+0.4358974358974359j
# beta_1 = 0.31614683863635784+0.47368421052631576j
# beta_1 = 0.21842105263157896+0.67j #.25+ .55j

# step_size = .0 + .1j  #.5*(1.0 + .5j)
# beta_1 = .1 + .5j

step_size = 1+0j#1 + 1j
step_size *= .5  # Play with this scale?
beta_1 = .25j  # Play with this beta?
# Can also play with the gradient.  Ex., [-y, x] -> [-a*y, b*x], a>0, b>0
# Can play with initialization


ratio, ws = experiment(step_size, beta_1, do_plot=True)
print(f"Final point: {ws[-1]}")
print(f"Rate of convergence: {ratio}")
print(f"Beta mag: {np.sqrt(np.imag(beta_1) ** 2 + np.real(beta_1) ** 2)}")
print(f"Negative momentum value: {-(np.imag(step_size) * np.imag(beta_1)) + np.real(beta_1)}")
print(f"Eta: {np.real(step_size)}")
plt.show()
#
# best_ratio = 1e16
# do_plot = True
# if do_plot:
#   fig, ax = plt.subplots()
# scale = 1.1
# num = 15 + 1
# for step_size_r in np.linspace(0.0, 2.0, num=10):  # (.75, 1.25, num=10):
#   for step_size_i in np.linspace(0.0, 2.0, num=10):  # (2.0, 4.0, num=20):
#     for beta_i in np.linspace(.0, 1.0, num=10):  # (0, .5, num=20):
#       beta_r_ratios = []
#       beta_r_domain = np.linspace(.0, np.sqrt(1.0 ** 2 - beta_i ** 2), num=10)  # (.0, np.sqrt(1.0 - beta_i**2), num=20)
#       best_ratio_flag = True  # False
#       for beta_r in beta_r_domain:
#         step_size = step_size_r + step_size_i * 1j
#         beta_1 = beta_r + beta_i * 1j
#         ratio, _ = experiment(step_size, beta_1)
#         beta_r_ratios += [ratio]
#         if ratio < best_ratio:
#           best_ratio = ratio
#           best_step_size = step_size
#           best_beta_1 = beta_1
#           print(f"Best ratio: {best_ratio} with step size = {step_size} and beta_1 = {beta_1}")
#           best_ratio_flag = True
#
#       if do_plot and best_ratio_flag:
#         ax.plot(beta_r_domain, beta_r_ratios, label=f'step = {step_size}, beta_i = {beta_i}')
#         ax.set_xlabel('beta_r')
#         ax.set_ylabel('convergence rate')
#         ax.set_ylim(-.1, 1.25)
#         ax.axhline(.679)
#         plt.legend(loc='upper right', bbox_to_anchor=(2.5, 1.0))
#
# # IDEA: boost im(alpha)*im(beta) to get more negative.  Boost re(beta) and im(alpha) to get longer history and same negative
