# -*- coding: utf-8 -*-
"""negative_momentum.ipynb

Automatically generated by Colaboratory.

Original file is located at
    https://colab.research.google.com/drive/11HzD0zcepA_XcM-Rm_-gagSc3cVwTESQ
"""

import numpy as np
#import jax.numpy as np
import matplotlib.pyplot as plt

def compute_loss(x, y):
  return x*y

def compute_gradient(x, y):
  return np.array([-y, x])
  #return np.array([2*x, 2*y])

def experiment(step_size, beta_1,
               beta_2=.99*np.exp(0*1j*np.pi), epsilon=1e-8*np.exp(0),
               do_plot=False):

  num_iterations = 500

  m = np.array([0 + 0j, 0 + 0j])
  v = np.array([0 + 0j, 0 + 0j])
  x_0, y_0 = 1, 1
  w = np.array([x_0, y_0])
  ws = [w]
  count = 0
  do_synthetic = True
  do_alternating = False
  for t in range(1, num_iterations):
      if do_synthetic:
        synthetic_g = compute_gradient(w[0], w[1])
        if len(ws) > 1:
          #print(beta_1*(w - ws[-2]))
          synthetic_w = w - step_size*synthetic_g + beta_1*(w - ws[-2])
        else:
          synthetic_w = w - step_size*synthetic_g
      else:
        synthetic_w = w

      g = compute_gradient(synthetic_w[0], synthetic_w[1])
      if do_alternating:
        player_num = t % 2
        if len(ws) > 1:
          w = w - step_size*np.array([g[0], 0]) + beta_1*np.array([(w[0] - ws[-2][0]), 0])
          g = compute_gradient(w[0], w[1])
          w = w - step_size*np.array([0, g[1]]) + beta_1*np.array([0, (w[1] - ws[-2][1])])
        else:
          w = w - step_size*np.array([g[0], 0])
          g = compute_gradient(w[0], w[1])
          w = w - step_size*np.array([0, g[1]])
      else:
        if len(ws) > 1:
          #print(beta_1*(w - ws[-2]))
          w = w - step_size*g + beta_1*(w - ws[-2])
        else:
          w = w - step_size*g
      ws += [np.real(w)]  # The imag component is 0, but this gets rid of the imag axis
      if ws[-1][0] > 3 or ws[-1][0] < -3 or ws[-1][1] > 3 or ws[-1][1] < -3:
        break
      #print(f"w = {w}\t combined_update = {combined_update}\t update = {update}\t update_conj = {update_conj}\tv_hat = {v_hat}\t m_hat = {m_hat}\t v = {v}\t m = {m}\t g = {g}")


  ws = np.array(ws)

  ratios = []
  norm = None
  for i in range(0, len(ws)):
    diff = ws[i] - [0, 0]
    old_norm = norm
    norm = diff @ diff.T
    if i > 0:
      ratio = norm / old_norm
      ratios += [ratio]

  if do_plot:
    fig, ax = plt.subplots()

    ax.scatter([x_0], [y_0], label='Init', c='b')
    start_trajectory = 0
    ax.plot(ws[start_trajectory:, 0],
            ws[start_trajectory:, 1], 
            label='Trajectory', alpha=.5, c='b')
    ax.scatter([0], [0], label='Solution', c='r')

    circle1 = plt.Circle((0, 0), 1.0, color='g', alpha=.25)
    ax.add_artist(circle1)

    ax.set_xlim([-3, 3])
    ax.set_ylim([-3, 3])
    plt.legend()

    fig, ax = plt.subplots()
    ax.plot(range(1, len(ws)), ratios)
    ax.plot(range(1, len(ws)), [np.mean(ratios[np.maximum(i - 100, 0):i]) for i in range(1, len(ws))])
    ax.set_ylim([.0, 1.5])
    ratios = np.array(ratios)
  return np.mean(ratios[-100:-1]), ws

#EG only
#step_size = 0.7067067067067067#1.0
#beta_1 = .0#-.1
# rate = .75

# EG + negative momentum
#step_size = 0.9494949494949496
#beta_1 = -0.1717171717171716
step_size = 1.1111111111111112
beta_1 = -0.1717171717171716
# rate = .27

# negative momentum + alternating
#step_size = 1.4141414141414144
#beta_1 = -0.21212121212121204
#step_size = 1.5555555555555556
#beta_1 = -0.1515151515151515

# A step size for multiply x*2
#step_size = 0.45845845845845845
#beta_1 = 0

# A step size for multiple x by 2
#step_size = 0.69
#beta_1 = -0.33

# A step siize for multiply x by 1/2
#step_size = 1.5757575757575757
#beta_1 = -0.1717171717171716

ratio, ws = experiment(step_size, beta_1, do_plot=True)
print(f"Final point: {ws[-1]}")
print(f"Rate of convergence: {ratio}")
print(f"Beta mag: {np.sqrt(np.imag(beta_1)**2 + np.real(beta_1)**2)}")
print(f"Negative momentum value: {-(np.imag(step_size) * np.imag(beta_1)) + np.real(beta_1)}")
print(f"Eta: {np.real(step_size)}")

best_ratio = 1e16
do_plot = False
if do_plot:
  fig, ax = plt.subplots()
scale = 1.1
num = 15 + 1
for step_size_r in np.linspace(1.0, 2.0, num=100):#(.75, 1.25, num=10):
  #for step_size_i in np.linspace(0.1, 2.0):#(2.0, 4.0, num=20):  
  #for beta_i in np.linspace(.0, 1.0, num=20):#(0, .5, num=20):
  beta_r_ratios = []
  beta_r_domain = np.linspace(-1.0, 0.0, num=100)
  best_ratio_flag = False
  for beta_r in beta_r_domain:
    step_size = step_size_r #+ step_size_i*1j
    beta_1 = beta_r #+ beta_i * 1j
    ratio, _ = experiment(step_size, beta_1)
    beta_r_ratios += [ratio]
    if ratio < best_ratio:
      best_ratio = ratio
      best_step_size = step_size
      best_beta_1 = beta_1
      print(f"Best ratio: {best_ratio} with step size = {step_size} and beta_1 = {beta_1}")
      best_ratio_flag = True
  
  
  if do_plot and best_ratio_flag:
    ax.plot(beta_r_domain, beta_r_ratios, label=f'step = {step_size}, beta_i = {beta_i}')
    ax.set_xlabel('beta_r')
    ax.set_ylabel('convergence rate')
    ax.set_ylim(0.25, 1.25)
    ax.axhline(.679)
    plt.legend(loc='upper right', bbox_to_anchor=(2.5, 1.0))

# IDEA: boost im(alpha)*im(beta) to get more negative.  Boost re(beta) and im(alpha) to get longer history and same negative

